#include <jni.h>
#include <string.h>
#include <stdio.h>
#include "JNIUtility.h"


/*
* input:
*	env:	the java-environment
*	jarr:	the java-array
* output:
*	T:	the dmatrix
* return:
*	0 if not succesfull
*	number of nodes otherwise
*/
int copyJavaToCint1D(env, jarr, T)
JNIEnv *env;
jintArray jarr;
int *T;
{
	// T is a vector with the first index starting at 1!!
	int i,imax;
	jint *carr;
	imax = arrayLength(env, jarr);
	
	carr = (*env)->GetIntArrayElements(env, jarr, NULL);

	if (carr == NULL)
	{
        	return 0; /* exception occurred */
	}

	for (i=0; i<imax; i++)
		T[i+1] = carr[i];

	(*env)->ReleaseIntArrayElements(env, jarr, carr, 0);
	return 1;
}


/*
* input:
*	env:	the java-environment
*	jarr:	the java-array
* output:
*	T:	the dmatrix
* return:
*	0 if not succesfull
*	number of nodes otherwise
*/
int copyJavaToCdouble1D(env, jarr, T)
JNIEnv *env;
jdoubleArray jarr;
double *T;
{
	// T is a vector with the first index starting at 1!!
	int i,imax;
	jdouble *carr;
	imax = arrayLength(env, jarr);
	
	carr = (*env)->GetDoubleArrayElements(env, jarr, NULL);
	if (carr == NULL)
	{
        	return 0; /* exception occurred */
	}

	for (i=0; i<imax; i++){
		T[i+1] = carr[i];
	}
	
	(*env)->ReleaseDoubleArrayElements(env, jarr, carr, 0);
	return 1;
}

int arrayLength(env, jarr)
JNIEnv *env;
jobjectArray jarr;
{
	int i;
	i = (int) (*env)->GetArrayLength(env, jarr);
	return i;
}

int ArrayIndexOutOfBoundsException(env,msg)
JNIEnv *env;
const char *msg;
{
	char s1[MAX_MSG];
	char s2[] = "Index has exceeded matrixdimensions. ";
	strcpy(s1,s2);
	strncat(s1, msg, ((int)(MAX_MSG-(int)strlen(msg)-(int)strlen(s2)) > 0)?(strlen(msg)):(MAX_MSG-(int)strlen(s2)));
	return throwException(env, "java/lang/ArrayIndexOutOfBoundsException",s1);
}

int Exception(env, msg)
JNIEnv *env;
const char *msg;
{
	return throwException(env, "java/lang/Exception", msg);
}

int OutOfMemoryError(env,msg)
JNIEnv *env;
const char *msg;
{
	char s1[MAX_MSG];
	char s2[] = "Not enough memory to complete operation. ";
	strcpy(s1,s2);
	strncat(s1, msg, ((int)(MAX_MSG-(int)strlen(msg)-(int)strlen(s2)) > 0)?(strlen(msg)):(MAX_MSG-(int)strlen(s2)));
	return throwException(env, "java/lang/OutOfMemoryError",s1);
}

int AddDBException(env,msg)
JNIEnv *env;
const char *msg;
{
	char s1[MAX_MSG];
	char s2[] = "Topology database not correct. ";
	strcpy(s1,s2);
	//printf("(int)(MAX_MSG-(int)strlen(msg)-(int)strlen(s2) > 0)?(strlen(msg)):(MAX_MSG-(int)strlen(s2))=%d\n",((int)(MAX_MSG-(int)strlen(msg)-(int)strlen(s2)) > 0)?(strlen(msg)):(MAX_MSG-(int)strlen(s2)));
	strncat(s1, msg, ((int)(MAX_MSG-(int)strlen(msg)-(int)strlen(s2)) > 0)?(strlen(msg)):(MAX_MSG-(int)strlen(s2)));
	return throwException(env, "be/ac/ulg/montefiore/run/totem/repository/model/exception/AddDBException",s1);
}

int NoRouteToHostException(env,msg)
JNIEnv *env;
const char *msg;
{
	char s1[MAX_MSG];
	char s2[] = "No route to host found. ";
	strcpy(s1,s2);
	strncat(s1, msg, ((int)(MAX_MSG-(int)strlen(msg)-(int)strlen(s2)) > 0)?(strlen(msg)):(MAX_MSG-(int)strlen(s2)));
	return throwException(env, "be/ac/ulg/montefiore/run/totem/repository/model/exception/NoRouteToHostException",s1);
}

int RoutingException(env,msg)
JNIEnv *env;
const char *msg;
{
	char s1[MAX_MSG];
	char s2[] = "Unknown exception thrown while routing. ";
	strcpy(s1,s2);
	strncat(s1, msg, ((int)(MAX_MSG-(int)strlen(msg)-(int)strlen(s2)) > 0)?(strlen(msg)):(MAX_MSG-(int)strlen(s2)));
	return throwException(env, "be/ac/ulg/montefiore/run/totem/repository/model/exception/RoutingException",s1);
}

int throwException(env, name, msg)
JNIEnv *env;
const char *name, *msg;
{
 	int i;
	jclass cls = (*env)->FindClass(env, name);
	/* If cls is NULL, an exception has already been thrown */
	if (cls != NULL)
	{
		i = (int) (*env)->ThrowNew(env, cls, msg);
	}
	else{
		fprintf(stderr,"Unable to find the exception class, giving up\n");
		return;
	}
	/* free the local ref */
	(*env)->DeleteLocalRef(env, cls);
	return i;
}
