/*
dijkstra_all_mod.c
This modules calculates the dijkstra-tree between a source and all other nodes
F.A. Kuipers 25/09/2003
*/


#include <string.h>
#include <stdio.h>
#include <math.h>
#include "nrutil.h"
#define NULL 0
#define POWER 10
#define INF 9.9e99

struct fiblist {	  /*global definition of the type fiblist, see Cormen: Fibonacci lists*/
	double key;
	int    id;
	int    el;
	int    degree;
	int    mark;	
	struct fiblist *parent;
	struct fiblist *child;
	struct fiblist *left;
	struct fiblist *right;
	};

struct el{
	struct fiblist *list;
	int index;
	struct el *next;
	};

void dijkstra_all(START,adj,max,datadj,aantalnodes,length)

int START;
int **adj;
int *max;
double **datadj;
int aantalnodes;
double *length;

{
  int countdown;
  int i,j,l,m,n,k,a,u,v,q,r,s,p,*counter,num,*previ,allextracted,*extracted;
  double *sum;
  struct fiblist *helppo, *heap, **heapnode;
  struct el **N;
  
  /*declaration of the functions maintaining the Fibonacci heap, see Cormen for implementing and maintaining Fibonacci heaps*/
  void create_heap();  			
  struct fiblist *extract_min_from_list();
  struct fiblist *insert();
  void consolidate();
  void decrease_key();
  void cut();
  void cascading_cut();
  void putlist();
  void delete_list();
  void insertlist();
  struct fiblist *getlist();
  
  /*memory allocation*/

  counter = ivector(1,aantalnodes);
  previ = ivector(1,aantalnodes);    
  sum = dvector(1,aantalnodes); 
  N = ivector(1,aantalnodes);
  heapnode = ivector(1,aantalnodes);
  extracted = ivector(1,aantalnodes);

  for(i=1;i<=aantalnodes;i++){
    N[i]=NULL;
    counter[i]=0;
    sum[i]=INF;
    length[i]=0.0;
    extracted[i]=0;
  }

  create_heap(&heap,&num);
  helppo=insert(&heap,&num,0.0,START,1);
  insertlist(&N[START],0,helppo);
  heapnode[START]=helppo;
  counter[START]+=1;
  sum[START]=0.0;
  
  countdown=0;  /*because only N-1 connections are measured, but in this case also END is taken into account*/
  allextracted = 0;

  /*main algorithm*/

  while(countdown < 1)
    {
      /*
	Select the minimum element from the Fibonacci list
      */
      helppo=extract_min_from_list(&heap,&num);  

      if(helppo==NULL)     
	{		
	  countdown=1;
	}      
      else
	{
	  i=helppo->id; /*i represents the extracted node from the fib heap*/

	  length[i]=helppo->key;

	  if(extracted[i]==0){
	    extracted[i]=1;
	    allextracted++;
	  }
	  if(allextracted==aantalnodes)
	    countdown=1;
	  
	  for(m=1; m<=max[i]; m++)
	    {
	      /*Go to the mth neighbour of node i*/
	      a=adj[i][m];
	      
	      if(sum[i]+datadj[i][m]<sum[a]){
		if(sum[a]==INF){
		  sum[a]=sum[i]+datadj[i][m];
		  helppo=insert(&heap,&num,sum[a],a,1);
		  heapnode[a]=helppo;
		  counter[a]+=1;
		  insertlist(&N[a],counter[a]-1,helppo);
		}
		else{
		  sum[a]=sum[i]+datadj[i][m];
		  decrease_key(&heap,heapnode[a],sum[i]+datadj[i][m]);
		}
		previ[a]=i;
	      }
	    }
	}
    }/*while*/

  /*free memory*/
 
  for(a=1;a<=aantalnodes;a++){
    delete_list(&N[a]);
  }

  free_ivector(N,1,aantalnodes);
  free_ivector(counter,1,aantalnodes);
  free_ivector(previ,1,aantalnodes);
  free_dvector(sum,1,aantalnodes);
  free_ivector(heapnode,1,aantalnodes);
  free_ivector(extracted,1,aantalnodes);
}
