#include <jni.h>
#include <stdio.h>
#include "nl_tudelft_repository_XAMCRA_JNIXAMCRA.h"
#include "nl_tudelft_repository_XAMCRA_SAMCRA.h"
#include "nrutil.h"
#include "JNIUtility.h"
#include "path_list.h"

/**********INTERFACES***************/

jintArray (*pJava_nl_tudelft_repository_XAMCRA_JNIXAMCRA_jnicomputePath)
  (JNIEnv *, jclass, jint, jint, jdouble, jdoubleArray, jint, jint);

/********END INTERFACES****************/


/********** GLOBAL VARIABLES ***************/
// Multi-Threading SAFE ???
int **adj;
int *numAdj;
double ***datAdj;
int *nodeMapping;

PathListStruct *pathList;
int N=0;		// Number of Nodes
int n=0;		// 
int numTotalLinks=0;	
int numTempLinks=0;
int M=0;		// Number of metrics
int semaphore = 0;
int qoSConstraintFlag = 0;
/**********END GLOBAL VARIABLES ***************/


int is_Samcra_ready()
{
	int result=0;
	if(n > 0 && N == n && numTempLinks > 0 && numTotalLinks == numTempLinks)
		result=1;
	return result;
}

int addPath(int *pathC, int numNodes, double cap)
{
	return insertData(pathList, pathC+1, numNodes, cap);
}

int removePath(int *pathC, int numNodes, double cap)
{
	return removeData(pathList, pathC+1, numNodes, cap);
}

void Java_nl_tudelft_repository_XAMCRA_SAMCRA_addRemovePath
  (JNIEnv *env, jclass cls, jintArray path, jdouble reservation, jint addRemoveFlag)
{
	// Add -> 0
	// Remove -> not 0
	if(!is_Samcra_ready() || !semaphore){
		AddDBException(env,"X-AMCRA is not ready... Did you call Init function? Was the topology database created properly?");
		return;
	}
	int i;
	int neighbourIndex, currentNode, nextNode;
	int *pathC, numNodes;
	int nodeMapped;
	
	if(addRemoveFlag && !length_list(pathList)){
		// Remove a path
		AddDBException(env,"Cannot remove a path since there is no path stablished.");
	}
	
	// Path is ok
	numNodes = arrayLength(env, path);
	if(numNodes < 2 || numNodes > N){
		AddDBException(env,"Check provided path.");
		return;
	}
	
	// Reservation is within the range
	if(reservation < 0 ){
		AddDBException(env,"Reservation: negative values not allowed...");
		return;
	}
	
	pathC = ivector(1,numNodes);
	i = copyJavaToCint1D(env,path, pathC);
	if (i==0){
		AddDBException(env,"Provided path could not be read.");
		return;
	}

	// Mapping
	for(i=1;i<=numNodes;i++){
		nodeMapped = 1;	
		while(nodeMapped <= N && pathC[i] != nodeMapping[nodeMapped]){
			nodeMapped++;
		}
		if(nodeMapped > N){
			AddDBException(env,"One of the nodes ID does not exist...");		// One of the nodes does not exist...
			return;
		}
		pathC[i] = nodeMapped;
	}	
	
	int *pathNeighbour = ivector(1,numNodes-1);
	// Checking of the existence of the path)
	for(i=1;i<numNodes;i++){ // The last node is not checked...
		// search the link
		neighbourIndex = 1;
		currentNode = pathC[i];
		nextNode = pathC[i+1];
		while(neighbourIndex <= numAdj[currentNode] && nextNode != adj[currentNode][neighbourIndex]){
			neighbourIndex++;
		}
		if(neighbourIndex > numAdj[currentNode]){
			AddDBException(env,"The node is not consistent...");		// The node is not a neighbour!!!!
			return;
		}
		pathNeighbour[i]=neighbourIndex;
		if(!addRemoveFlag){
			// Add a path
			if(datAdj[0][currentNode][neighbourIndex] < reservation){
				AddDBException(env,"Not enough resources available...");
				return;
			}
		}
	}
	
	// Memory Management & insert/remove paths
	if(!addRemoveFlag){
		// Add a path
		if(!addPath(pathC,numNodes,reservation)){
			AddDBException(env,"Could not add the path.");
			return;
		}
	}
	else{
		// Remove a path
		if(!removePath(pathC,numNodes,reservation)){
			AddDBException(env,"Could not remove path.");
			return;
		}	
	}	
	
	// make the reservation or release
	for(i=1;i<numNodes;i++){ // The last node is not checked...
		// search the link
		neighbourIndex = pathNeighbour[i];
		currentNode = pathC[i];
		
		if(!addRemoveFlag){
			// Add a path
			datAdj[0][currentNode][neighbourIndex] -= reservation;		
		}
		else{
			// Remove a path 
			datAdj[0][currentNode][neighbourIndex] += reservation;
		}
	}

	free_ivector(pathC,1,numNodes);
	free_ivector(pathNeighbour,1,numNodes-1);
	return;
}

/*
 * Class:     nl_tudelft_repository_XAMCRA_JNIXAMCRA
 * Method:    jniprintXamcraDB
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_nl_tudelft_repository_XAMCRA_JNIXAMCRA_jniprintXamcraDB
  (JNIEnv *env, jclass cls)
{
	int i,j,s;
	if(!is_Samcra_ready() || !semaphore){
		fprintf(stdout,"Topology Database is not ready\nComplete the DB and call again.\n");
		return;
	}
	fprintf(stdout,"############## TOPOLOGY ##############\n");
	for(i=1;i<=N;i++){
		fprintf(stdout,"\nnumadj[%d] = %d:",nodeMapping[i],numAdj[i]);
		for(j=1;j<=numAdj[i];j++){
			fprintf(stdout," %d([%.2lf] ",nodeMapping[adj[i][j]],datAdj[0][i][j]);
			for(s=1;s<=M;s++){
				fprintf(stdout," %.2lf",datAdj[s][i][j]);
			}
			fprintf(stdout,")");
		}
	}
	fprintf(stdout,"\n");
	fprintf(stdout,"############## ALLOCATED PATH LIST ##############\n");
	listData(pathList,nodeMapping);	
	
}

/*
 * Class:     nl_tudelft_repository_XAMCRA_JNIXAMCRA
 * Method:    jniinitXamcra
 * Signature: (IIII)V
public native static void jniinitXamcra(int numNodes, int numLinks, int nbMetrics, int xamcraVersion);
 */
JNIEXPORT void JNICALL Java_nl_tudelft_repository_XAMCRA_JNIXAMCRA_jniinitXamcra
(JNIEnv *env, jclass cls, jint numNodes, jint numLinks, jint nbMetrics, jint xamcraVersion)
{
    //printf("Hello Java_nl_tudelft_repository_XAMCRA_SAMCRA_initXamcra\n");
	int i;
	// Negative values accepted...
	
	if(numNodes < 2 || numLinks < 1 || nbMetrics < 1 || semaphore){
		// cannot throw an exception... Should I??? 
		printf("initXamcra: check input parameters. Initialization NOT done or trying to do twice without calling kill function.\n");
		return;
	}	
	
	switch (xamcraVersion) {
		case 0:
			semaphore = 1;
			//SAMCRA
			pJava_nl_tudelft_repository_XAMCRA_JNIXAMCRA_jnicomputePath = &Java_nl_tudelft_repository_XAMCRA_SAMCRA_computePath;			
			 break;
	 	default:
			semaphore = 0;
		 	//Version Unknown!!!;
		 	break;
	}
	if(!semaphore)
		return;

	N = numNodes;
	n = 0;
	numTotalLinks = numLinks;
	numTempLinks = 0;
	
	if(nbMetrics == 1){
		qoSConstraintFlag = 1;
		M = 1;
	}
	else{
		qoSConstraintFlag = 0;
		M = nbMetrics-1;		// nbMetrics includes Bw!!!
	}
	
	
	pathList = createList();
	if(pathList == NULL){
		printf("Structure to keep track of the allocated paths could not be initialised.\n");
		return;
	}
	
	if(numAdj != 0)
		free_ivector(numAdj,1,N);
	numAdj = ivector(1,N);
//	if(numAdj == 0)
//		OutOfMemoryError(env);
	for(i=1;i<=N;i++) numAdj[i]=0;
	if(adj != 0)
		free_imatrix(adj,1,N,1,N-1);
	adj = imatrix(1,N,1,N-1);
//	if(adj == 0)
//		OutOfMemoryError(env);

	if(datAdj != 0)
		free_d3tensor(datAdj,0,M,1,N,1,N-1);	
	datAdj = d3tensor(0,M,1,N,1,N-1);
//	if(datadj == 0)
//		OutOfMemoryError(env);

	if(nodeMapping != 0)
		free_ivector(nodeMapping,1,N);
	nodeMapping = ivector(1,N);
//	if(nodeMapping == 0)
//		OutOfMemoryError(env);
	
	return;
}


/*
 * Class:     nl_tudelft_repository_XAMCRA_JNIXAMCRA
 * Method:    jnikillXamcra
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_nl_tudelft_repository_XAMCRA_JNIXAMCRA_jnikillXamcra
  (JNIEnv *env, jclass cls)
{
    //printf("Hello Java_nl_tudelft_repository_XAMCRA_SAMCRA_killXamcra\n");
	if(!semaphore)
		return;	
	
	if(numAdj!=0)
		free_ivector(numAdj,1,N);
	numAdj = NULL;
	if(datAdj!=0)
		free_d3tensor(datAdj,0,M,1,N,1,N-1);	
	datAdj = NULL;
	if(adj!=0)
		free_imatrix(adj,1,N,1,N-1);
	adj = NULL;
	if(nodeMapping!=0)
		free_ivector(nodeMapping,1,N);
	nodeMapping = NULL;
	n = 0;N = 0;
	numTotalLinks = 0;numTempLinks=0;
	M = 0;
	free_List(pathList);
	pathList = NULL;
	semaphore = 0;
	qoSConstraintFlag = 0;
	return;
}

/*
 * Class:     nl_tudelft_repository_XAMCRA_JNIXAMCRA
 * Method:    jniaddNode
 * Signature: (I)V
public native static void jniaddNode(int nodeId) throws AddDBException;
 */
JNIEXPORT void JNICALL Java_nl_tudelft_repository_XAMCRA_JNIXAMCRA_jniaddNode
  (JNIEnv *env, jclass cls, jint nodeId)
{
    //printf("Hello Java_nl_tudelft_repository_XAMCRA_SAMCRA_addNode\n");	
	if(is_Samcra_ready() || n<0 || n>=N || !semaphore){
		AddDBException(env,"Check input parameters. Initialization might NOT been done.");
		return;
	}
	
	nodeMapping[++n]=nodeId;
	return;
}


/*
 * Class:     nl_tudelft_repository_XAMCRA_JNIXAMCRA
 * Method:    jniaddLink
 * Signature: (IID[D)V
public native static void jniaddLink(int srcId, int dstId, double cap, double[]  qosMetric) throws AddDBException;
 */
JNIEXPORT void JNICALL Java_nl_tudelft_repository_XAMCRA_JNIXAMCRA_jniaddLink
  (JNIEnv *env, jclass cls, jint srcId, jint dstId, jdouble cap, jdoubleArray qosMetrics)
{
    //printf("Hello Java_nl_tudelft_repository_XAMCRA_SAMCRA_addLink\n");
	if(is_Samcra_ready() || numTempLinks<0 || numTempLinks >= numTotalLinks || !semaphore){
		AddDBException(env,"Check input parameters. Initialization might NOT been done.");
		return;
	}

	int i;
	double *qosMetricsC;

	if(srcId == dstId){
		AddDBException(env,"Source and Destination nodes should NOT be the same.");
		return;
	}
	
	// cap is in the range
	if(cap < 0 ){
		AddDBException(env,"Negative values of capacity not allowed.");
		return;
	}
	
	// Both nodes exist
	int srcNodeMapped, dstNodeMapped;
	srcNodeMapped = 1;	
	while(srcNodeMapped <= n && srcId != nodeMapping[srcNodeMapped]){
		srcNodeMapped++;
	}
	if(srcNodeMapped > n){
		AddDBException(env,"Source node ID not correct.");
		return;
	}
	
	dstNodeMapped = 1;	
	while(dstNodeMapped <= n && dstId != nodeMapping[dstNodeMapped]){
		dstNodeMapped++;
	}
	if(dstNodeMapped > n){
		AddDBException(env,"Destination node ID not correct.");
		return;
	}
	
	// The link is not already added
	i = 1;
	while(i <= numAdj[srcNodeMapped] && dstNodeMapped != adj[srcNodeMapped][i]){
		i++;
	}
	
	if(i<= numAdj[srcNodeMapped])
	{
		AddDBException(env,"The link already exists.");		//The link already exists. 
		return;
	}	
	
	// qosmetrics are ok
	if(qoSConstraintFlag){
		qosMetricsC = dvector(1,M);			// M = 1
		qosMetricsC[1] = 0;
	}
	else{
		i = arrayLength(env, qosMetrics);
		if(i != M){
			AddDBException(env,"An unknown problem occurred with QoS metrics.");
			return;
		}
		
		qosMetricsC = dvector(1,M);
		i = copyJavaToCdouble1D(env, qosMetrics, qosMetricsC);
		if (i==0){
			AddDBException(env,"An unknown problem occurred with QoS metrics.");
			return;
		}
	}

	// Add Link
	adj[srcNodeMapped][++numAdj[srcNodeMapped]] = dstNodeMapped;
	for(i=1;i<=M;i++){
		datAdj[i][srcNodeMapped][numAdj[srcNodeMapped]]=qosMetricsC[i];
	}
	datAdj[0][srcNodeMapped][numAdj[srcNodeMapped]]=cap;
	numTempLinks++;
	free_dvector(qosMetricsC,1,M);
	return;
}


/*
 * Class:     nl_tudelft_repository_XAMCRA_JNIXAMCRA
 * Method:    jniaddPath
 * Signature: ([ID)V
public native static void  jniaddPath(int[] path, double reservation) throws AddDBException;
 */
JNIEXPORT void JNICALL Java_nl_tudelft_repository_XAMCRA_JNIXAMCRA_jniaddPath
  (JNIEnv *env, jclass cls, jintArray path, jdouble reservation)
{
    //printf("Hello Java_nl_tudelft_repository_XAMCRA_JNIXAMCRA_jniaddPath\n");
	return Java_nl_tudelft_repository_XAMCRA_SAMCRA_addRemovePath(env,cls,path,reservation,0);
}


/*
 * Class:     nl_tudelft_repository_XAMCRA_JNIXAMCRA
 * Method:    jniremovePath
 * Signature: ([ID)V
public native static void  jniremovePath(int[] path, double reservation) throws AddDBException;
 */
JNIEXPORT void JNICALL Java_nl_tudelft_repository_XAMCRA_JNIXAMCRA_jniremovePath
  (JNIEnv *env, jclass cls, jintArray path, jdouble reservation)
{
    //printf("Hello Java_nl_tudelft_repository_XAMCRA_JNIXAMCRA_jniremovePath\n");
	return Java_nl_tudelft_repository_XAMCRA_SAMCRA_addRemovePath(env,cls,path,reservation,1);
}


/*
 * Class:     nl_tudelft_repository_XAMCRA_JNIXAMCRA
 * Method:    jnicomputePath
public native static int[] jnicomputePath(int src, int dst, double bandwidth, double[] qosConstraint, int ADDLSP) throws AddDBException, NoRouteToHostException, RoutingException;
 * Signature: (IID[DI)[I
 */
JNIEXPORT jintArray JNICALL Java_nl_tudelft_repository_XAMCRA_JNIXAMCRA_jnicomputePath
  (JNIEnv *env, jclass cls, jint src, jint dst, jdouble bandwidth, jdoubleArray qosConstraints, jint ADDLSP)
{
	return pJava_nl_tudelft_repository_XAMCRA_JNIXAMCRA_jnicomputePath(env,cls,src,dst,bandwidth,qosConstraints,qoSConstraintFlag,ADDLSP);	
}
