#include <jni.h>
#include <stdio.h>
#include "nrutil.h"
#include "JNIUtility.h"
#include "samcra_la_mod.h"

// Multi-Threading problems ???
extern int **adj;
extern int *numAdj;
extern double ***datAdj;
extern int *nodeMapping;

// Multi-Threading problems ???
extern int N;		// Number of Nodes	
extern int M;		// Number of metrics
extern int semaphore;

void Java_nl_tudelft_repository_XAMCRA_SAMCRA_addRemovePath
  (JNIEnv *env, jclass cls, jintArray path, jdouble reservation, jint addRemoveFlag);

/*
 * Class:     nl_tudelft_repository_XAMCRA_SAMCRA
 * Method:    jnicomputePath
public native static int[] jnicomputePath(int src, int dst, double bandwidth, double[] qosConstraint, int ADDLSP) throws AddDBException, NoRouteToHostException, RoutingException;
 * Signature: (IID[DI)[I
 */
jintArray Java_nl_tudelft_repository_XAMCRA_SAMCRA_computePath
  (JNIEnv *env, jclass cls, jint src, jint dst, jdouble bandwidth, jdoubleArray qosConstraints, jint qoSConstraintFlag, jint ADDLSP)
{
    //printf("Hello Java_nl_tudelft_repository_XAMCRA_SAMCRA_computePath\n");	
	if(!is_Samcra_ready() || !semaphore){
		AddDBException(env,"Check input parameters. Initialization might NOT been done.");
		return;
	}
	
	int i,j,t,*numAdjLocal, **adjLocal, minimum_k, k_used, s;
	int *returnedPathC;
	double ***datAdjLocal;
	double *qosMetricsC,*weights;
	int hops;

	if(src == dst){
		AddDBException(env,"Source and Destination nodes should NOT be the same.");
		return;
	}
	
	// Mapping of nodes
	int srcNodeMapped, dstNodeMapped;
	srcNodeMapped = 1;	
	while(srcNodeMapped <= N && src != nodeMapping[srcNodeMapped]){
		srcNodeMapped++;
	}
	if(srcNodeMapped > N){
		RoutingException(env,"Source node ID not correct.");
		return;
	}

	dstNodeMapped = 1;	
	while(dstNodeMapped <= N && dst != nodeMapping[dstNodeMapped]){
		dstNodeMapped++;
	}
	
	if(dstNodeMapped > N){
		RoutingException(env,"Destination node ID not correct.");
		return;
	}

	if(qoSConstraintFlag){
		qosMetricsC = dvector(1,M);			// M = 1
		qosMetricsC[1] = 1;
	}
	else{
		// qosmetrics are ok??
		i = arrayLength(env, qosConstraints);
		if(i != M){
			RoutingException(env,"An unknown problem occurred with QoS metrics.");
			return;
		}
		qosMetricsC = dvector(1,M);
		i = copyJavaToCdouble1D(env, qosConstraints, qosMetricsC);
		if (i==0){
			RoutingException(env,"An unknown problem occurred with QoS metrics.");
			return;
		}
	}

	returnedPathC = ivector(1,N);
	numAdjLocal = ivector(1,N);
	adjLocal = imatrix(1,N,1,N-1);
	datAdjLocal = d3tensor(1,M,1,N,1,N-1); // Bw not included!!!
	weights = dvector(1,M);

	// Prune the new topology
	for(i=1;i<=N;i++){
		numAdjLocal[i] = 0;
		for(j=1;j<=numAdj[i];j++){
			if(datAdj[0][i][j] >= bandwidth){
				numAdjLocal[i]++;
				adjLocal[i][numAdjLocal[i]] = adj[i][j];
				for(t=1;t<=M;t++){
					datAdjLocal[t][i][numAdjLocal[i]] = datAdj[t][i][j];
				}				
			}	
		}
	}
	
	hops = 0;
	samcrala(1,srcNodeMapped, dstNodeMapped, adjLocal, numAdjLocal, datAdjLocal, M, qosMetricsC, N, returnedPathC, &hops, weights, &minimum_k, &k_used);
	if(hops==0){
		NoRouteToHostException(env,"");
		return;
	}

	jint size = hops+1;	/* the number of elements */

	s = (int)(size)/2;
	// Back - Mapping and reverse the vector
	for(i=1; i<=s; i++){
		t = nodeMapping[returnedPathC[size - i + 1]];
		returnedPathC[size - i + 1] = nodeMapping[returnedPathC[i]];
		returnedPathC[i] = t;
	}
	if(size%2 == 1)	returnedPathC[s+1] = nodeMapping[returnedPathC[s+1]];
	
	jintArray int1DArr;

	int1DArr = (*env)->NewIntArray(env, size);
	if (int1DArr == NULL){
		RoutingException(env,"An unknown problem allocating memory...");
		return;
	}

	(*env)->SetIntArrayRegion(env, int1DArr, 0, size, returnedPathC+1);
	
	if(ADDLSP){
		// Add the path
		Java_nl_tudelft_repository_XAMCRA_SAMCRA_addRemovePath(env, cls, int1DArr, bandwidth,0);
	}	

	free_ivector(numAdjLocal,1,N);
	free_d3tensor(datAdjLocal,1,M,1,N,1,N-1);
	free_imatrix(adjLocal,1,N,1,N-1);
	free_ivector(qosMetricsC,1,M);
	free_ivector(returnedPathC,1,N);
	free_dvector(weights,1,M);	
	return int1DArr;
}


