/*This module is created to simulate a dynamic array and to make/alter linked lists
Eguzki Astiz 6/12/2005*/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "path_list.h"

unsigned int counter = 0;

typedef struct PathStruct {
	unsigned int *nodes;
	double bw; 
	unsigned long int length;
} Path;

typedef struct dynamic_array PathListItem;

struct dynamic_array{
	Path path;						/*data gives the data in the array*/
	unsigned int id;                  	
	PathListItem *next;  			/*point to the next element in the array*/
};

struct ListStruct{              
  unsigned int numPaths;         
  PathListItem* ptr; 
};

PathListStruct *createList()
{
	PathListStruct *pathListStruct = NULL;
	pathListStruct = (PathListStruct*)malloc(sizeof(PathListStruct));
	if(pathListStruct != NULL){
		pathListStruct->numPaths = 0;
		pathListStruct->ptr = NULL;
	}
	return pathListStruct;
}

/*This function calculates the length of a dynamic_array*/
int length_list(PathListStruct *str)
{
	if(str == NULL){
		return -1;
	}
	return str->numPaths;
}

void free_List(PathListStruct *str)
{
	if(str == NULL){
		return;
	}
	int tempNumPaths;
	PathListItem *item,*temp;
	tempNumPaths = str->numPaths;
	item = str->ptr;
	while(item != NULL && tempNumPaths){
		temp = item;
		item = item->next;
		free(temp->path.nodes);
		free(temp);
		tempNumPaths--;
	}
	str->numPaths = 0;
	return;	
}

int removeData(PathListStruct *str, int *path, int numNodes, double cap)
{
	if(str == NULL || !length_list(str) || numNodes < 2){
		return 0;
	}
	
	PathListItem *item = NULL;
	PathListItem *prevItem = NULL;
	int found = 0;
	int j,i;
	
	i=0;
	item = str->ptr;
	do{
		if(item->path.length == numNodes && item->path.bw == cap){
			j=0;
			while(j<numNodes && path[j] == item->path.nodes[j]) j++;
			if(j==numNodes) found = 1;
		}
		i++;
		if(found) break;
		
		prevItem = item;
		item = item->next;	
	}while(i<str->numPaths && item!=NULL);
	
	if(found){
		if(prevItem == NULL){
			str->ptr = item->next;
		}
		else{
			prevItem->next = item->next;
		}
		free(item->path.nodes);
		free(item);
		str->numPaths--;
	}	
	return found;		
}

int insertData(PathListStruct *str, int *path, int numNodes, double cap)
{
	int i;
	if(str == NULL){
		return 0;
	}
	
	PathListItem *item = NULL;
	
	item = (PathListItem*)malloc(sizeof(PathListItem));
	if(item == NULL)
		return 0;
	item->path.nodes = (unsigned int*)malloc(numNodes*sizeof(unsigned int));
	
	if(item->path.nodes == NULL){
		free(item);
		return 0;
	}
	
	for(i=0;i<numNodes;i++){
		item->path.nodes[i] = path[i];
	}
	
	item->path.length=numNodes;
	item->next = str->ptr;
	counter = (counter+1)%(UINT_MAX);
	item->id = counter;
	item->path.bw = cap;
	str->ptr = item;	
	str->numPaths++;
	return 1;
}

void listData(PathListStruct *str, int *nodeMapping){
	PathListItem *item = NULL;
	int i, tempNumPaths;
	
	if(str == NULL){
		return 0;
	}
	
	if(!str->numPaths){
		printf("There are no paths allocated.\n");
		return;
	}
	
	printf("%d paths allocated.\n",str->numPaths);
	item = str->ptr;
	tempNumPaths = str->numPaths;
	while(item != NULL && tempNumPaths){
		printf("PathID %u[%.2lf]:",item->id,item->path.bw);
		for(i=0;i<item->path.length; i++){
			printf("%d\t",nodeMapping[item->path.nodes[i]]);
		}
		printf("\n");
		item = item->next;
		tempNumPaths--;
	}
}


