/*
samcra_la_mod.c
This modules calculates the samcra-path between a source and destination node
It makes use of dijkstra as look-ahead information
F.A. Kuipers 27/10/2003
*/


#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "nrutil.h"
#include "linked_list.h"
#include "samcra_la_mod.h"
#define NULL 0
#define POWER 10
#define INF 9.9e99
#define TINY 10e-10

struct fiblist {	  /*global definition of the type fiblist, see Cormen: Fibonacci lists*/
	double key;
	int    id;
	int    el;
	int    degree;
	int    mark;	
	struct fiblist *parent;
	struct fiblist *child;
	struct fiblist *left;
	struct fiblist *right;
	};

struct el{
	struct fiblist *list;
	int index;
	struct el *next;
	};

struct dynamic_array{
  int data1;
 int index1;
  struct dynamic_array *next1;
};

struct dynamic_vector{
  double data2;
  int index2;
  struct dynamic_vector *next2;
};


void samcrala(MCP,START,END,adj,max,datadj,numm,A,aantalnodes,path,hopcount,weights,minimum_k,k_used)

int MCP;
int START;
int END;
int **adj;
int *max;
double ***datadj;
int numm;
int aantalnodes;
double *A;
int *path;
int *hopcount;
double *weights;
int *minimum_k;
int *k_used;

{
  int countdown,mark,*prevnodes,equal,dom,olddom;
  int i,j,l,m,n,k,a,u,v,q,r,s,p,*counter,num,pathcounter,*max_q;
  double *intermediate, endvalue,prediction,newvalue;
  double **lowerbounds,length;
  struct dynamic_array **previ, **prevj, **poscounter,**marked,**extracted;
  struct dynamic_vector ***sum;
  struct fiblist *helppo, *heap;
  struct el **N;
  
  /*declaration of the functions maintaining the Fibonacci heap, see Cormen for implementing and maintaining Fibonacci heaps*/
  void create_heap();  			
  struct fiblist *extract_min_from_list();
  struct fiblist *insert();
  void consolidate();
  void decrease_key();
  void cut();
  void cascading_cut();
  void putlist();
  void delete_list();
  void insertlist();
  struct fiblist *getlist();
  
  /*memory allocation*/

  counter = ivector(1,aantalnodes);
  intermediate = dvector(1,numm);
  previ = ivector(1,aantalnodes);      /*previous node in the path*/
  prevj = ivector(1,aantalnodes);      /*index of previous node*/
  prevnodes = ivector(1,aantalnodes);
  poscounter = ivector(1,aantalnodes);
  sum = dmatrix(1,numm,1,aantalnodes);  /*keeps track of the sum of the linkmetrics on the path*/
  marked = ivector(1,aantalnodes);
  //extracted = ivector(1,aantalnodes);
  N = ivector(1,aantalnodes);
  max_q = ivector(1,aantalnodes);
  lowerbounds = dmatrix(1,numm,1,aantalnodes);

  endvalue = 1.0;
  pathcounter=0;
  (*hopcount)=0;
  (*minimum_k)=0;
  (*k_used) = 0;
  countdown=0;  /*because only N-1 connections are measured, but in this case also END is taken into account*/

  for(i=1;i<=aantalnodes;i++){
    previ[i]=NULL;
    prevj[i]=NULL;
    poscounter[i]=NULL;
    marked[i]=NULL;
    //extracted[i]=NULL;
    N[i]=NULL;
    counter[i]=0;
    max_q[i]=0;
    path[i]=0;
    prevnodes[i]=0;
  }
  
  for(i=1;i<=numm;i++){
    for(j=1;j<=aantalnodes;j++){
      sum[i][j]=NULL;
      lowerbounds[i][j]=0.0;
    }
  }
  
  for(i=1;i<=numm;i++){
    dijkstra_all(END,adj,max,datadj[i],aantalnodes,lowerbounds[i]);
    if(lowerbounds[i][START]>A[i]){
      countdown = 1;
      (*k_used)=1;
      i=numm+1;
    }
  }
  
  create_heap(&heap,&num);
  helppo=insert(&heap,&num,0.0,START,1);
  insertlist(&N[START],0,helppo);
  counter[START]+=1;
  for(l=1; l<=numm; l++){
    insertdatad(&sum[l][START],0,0.0);
  }
  
  /*for(i=1;i<=aantalnodes;i++)
  //for(j=1;j<=length_array(poscounter[i]);j++)
  insertdata(&poscounter[i],j-1,0);*/
  
  /*main algorithm*/

  while(countdown < 1)
    {
      /*
	Select the minimum element from the Fibonacci list
      */
      helppo=extract_min_from_list(&heap,&num);  
      
      if(helppo==NULL)     
	{		
	  countdown=1;
	}      
      else
	{
	  i=helppo->id; /*i represents the extracted node from the fib heap*/
	  j=helppo->el;

	  if(getdata(marked[i],j-1)!=1){
	    
	    /*if(i==START)
	      q=1;
	    else{
	      q=getdata(poscounter[getdata(previ[i],j-1)],(getdata(prevj[i],j-1)-1));
	    }
	    for(s=1;s<=length_array(poscounter[i]);s++){       
	      if(getdata(poscounter[i],s-1)==q){ 
		q++;
		s=0;
	      }
	    }
	    insertdata(&poscounter[i],j-1,q);
	    max_q[i]=q;*/
	    
	    if(i==END){
	      countdown++;
	      pathcounter=1;
	      /*for(l=1;l<=numm;l++)
		fprintf(stderr,"sam[%d] = %lf\n",l,getdatad(sum[l][i],j-1)/A[l]);*/
	      for(s=1; s<=numm; s++){
		weights[s] = getdatad(sum[s][i],j-1);
	      }
	      u=i;
	      v=j;
	      while(u!=START){

		//fprintf(stderr,"samcra path: %d\n",u);

		path[pathcounter++]=u;
		s=u;
		u = getdata(previ[u],v-1);
		v = getdata(prevj[s],v-1);
	      }
	      path[pathcounter]=u;
	      /*number of hops taken by the path as calculated by samcra = pathcounter*/
	      (*hopcount) = pathcounter-1;
	    }
	    
	    for(m=1; m<=max[i]; m++)
	      {
		/*Go to the mth neighbour of node i*/
		a=adj[i][m];
		length = 0.0;
		prediction = 0.0;
		//newvalue = 0.0;
		
		/*
		  intermediate[l] = sum of all the lth link weights up to node a
		  newvalue = maximum of intermediate[l] devided by the requested
		  lth constraint
		  prediction = predicted e2e-length, based on intermediate + lowerbounds
		*/
		for(l=1; l<=numm; l++)
		  {
		    intermediate[l]=(getdatad(sum[l][i],j-1)+datadj[l][i][m]);
		    /*if(intermediate[l]/A[l]>newvalue)
		      newvalue = intermediate[l]/A[l];*/
		    if((intermediate[l]+lowerbounds[l][a])/A[l]>prediction)
		      prediction = (intermediate[l]+lowerbounds[l][a])/A[l];
		    length += (intermediate[l]+lowerbounds[l][a])/A[l];
		  } /*for*/
		
                if(length>numm)
                  prediction  = endvalue+1; /*if length>constraintslength, the subpath cannot lead to a feasible*/
		else{
		  
		  p=1;
		  /*
		    Check if a is not dominated. This is checked against all entries
		    that have entered the queue of node a. If a is not dominated then
		    p=counter[a]+1. Counter[a] represents the number of times an entry
		    entered the queue of a.
		  */
		  while(p<=counter[a]){
		    dom=0;
		    olddom=0;
		    for(l=1;l<=numm;l++){
		      if((intermediate[l]-getdatad(sum[l][a],p-1)) >= 0.0)
			dom++;
		      if((intermediate[l]-getdatad(sum[l][a],p-1)) < 0.0)
			olddom++;
		    }
		    if(dom==numm)
		      p=(counter[a]+1);
		    else{ 
		      if((olddom==numm)||(getlist(N[a],p-1)->key>endvalue))
			insertdata(&marked[a],p-1,1);
		    }
		    p++;
		  }
		}
		  
		/*
		  Newvalue shouldn't exceed 1, because then the constraints aren't met.
		  The new entrie may also not be dominated (p=counter[a]+1). Then we
		  can insert the node. First we keep track of the path by inserting a's
		  previous node i and its place in the queue j. We also keep track of the
		  sum of the metrics up to node a. Then we insert the node into the fibonacci 
		  heap and increase the counter.
		*/
		if((prediction <= endvalue)&&(p==(counter[a]+1)))   
		  { 
		    for(p=1;p<=counter[a];p++){
		      if(getdata(marked[a],p-1)==1){
			if(prediction<=getlist(N[a],p-1)->key){
			  insertdata(&previ[a],p-1,i);
			  insertdata(&prevj[a],p-1,j);
			  for(l=1; l<=numm; l++){
			    insertdatad(&sum[l][a],p-1,intermediate[l]);
			  }
			  decrease_key(&heap,getlist(N[a],p-1),prediction);
			  //decrease_key(&heap,getlist(N[a],p-1),newvalue);
			  //decrease_key(&heap,getlist(N[a],p-1),length);
			  insertdata(&marked[a],p-1,0);
			  p=counter[a]+2;
			}
		      }
		    }

		    if(p==counter[a]+1){
		      insertdata(&previ[a],p-1,i);
		      
		      insertdata(&prevj[a],p-1,j);
		      for(l=1; l<=numm; l++){
			insertdatad(&sum[l][a],p-1,intermediate[l]);
		      }
		      helppo=insert(&heap,&num,prediction,a,p);
		      //helppo=insert(&heap,&num,newvalue,a,p);
		      //helppo=insert(&heap,&num,length,a,p);
		      counter[a]+=1;
		      insertlist(&N[a],counter[a]-1,helppo);
		    }
		    
		    if(a==END){
		      endvalue = prediction;
		      if(MCP==1){
			decrease_key(&heap,getlist(N[a],p-1),0.0);
		      }
		    }
		  }/*newvalue*/
		}/*for*/	  
	  }/*if not marked*/
	  else
	    insertdata(&marked[a],p-1,0);
	}/*else*/
    }/*while*/
  
  /*minimum queue-size needed to retrieve the shortest path
    (*minimum_k) = max_q[END];*/

  /*queue-size used to calculate the path*/
  for(a=1;a<=aantalnodes;a++){
    if(counter[a]>(*k_used))
      (*k_used) = counter[a];
  }
  
  /*free memory*/
 
  for(a=1;a<=aantalnodes;a++){
    delete_list(&N[a]);
  }

  for(a=1;a<=aantalnodes;a++){
    delete_array(&previ[a]);
    delete_array(&prevj[a]);
    delete_array(&poscounter[a]);
    delete_array(&marked[a]);
    //delete_array(&extracted[a]);
  }
  
  for(a=1;a<=numm;a++){
    for(i=1;i<=aantalnodes;i++){
      delete_vector(&sum[a][i]);
    }
  }

  free_ivector(N,1,aantalnodes);
  free_ivector(max_q,1,aantalnodes);
  free_ivector(counter,1,aantalnodes);
  free_ivector(previ,1,aantalnodes);
  free_ivector(prevj,1,aantalnodes);
  free_ivector(prevnodes,1,aantalnodes);
  free_ivector(poscounter,1,aantalnodes);
  free_ivector(marked,1,aantalnodes);
  //free_ivector(extracted,1,aantalnodes);
  free_dvector(intermediate,1,numm);
  free_dmatrix(sum,1,numm,1,aantalnodes);
  free_dmatrix(lowerbounds,1,numm,1,aantalnodes);
}
